
/******************************
 * forms
 ******************************/

/* note that this should only apply to "normal" primary forms */
/* TODO: replace this with bulma equivalent */
.form {
    padding-left: 5em;
}

.field.is-horizontal .field-label .label {
    text-align: left;
    white-space: nowrap;
    width: 18em;
}

/* note that this should only apply to "normal" primary forms */
.form-wrapper .form .field.is-horizontal .field-body {
    min-width: 30em;
}

/* note that this should only apply to "normal" primary forms */
.form-wrapper .form .field.is-horizontal .field-body .select,
.form-wrapper .form .field.is-horizontal .field-body .select select {
    width: 100%;
}
