
/******************************
 * Styles for purchases
 ******************************/

div.field-wrapper {
    padding: 0;
}

.field-wrapper.notes .field {
    line-height: 1.1em;
    white-space: pre;
}
