
/******************************
 * tweaks for root user
 ******************************/

.navbar .navbar-menu .navbar-link.root-user,
.navbar .navbar-menu .navbar-item.root-user,
.navbar.is-white .navbar-item.has-dropdown.is-active .navbar-link.root-user,
.navbar.is-white .navbar-item.has-dropdown:hover .navbar-link.root-user {
    background-color: red;
}
