# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Tasks for Tailbone
"""

from __future__ import unicode_literals, absolute_import

import os
import shutil

from invoke import task


here = os.path.abspath(os.path.dirname(__file__))
exec(open(os.path.join(here, 'tailbone', '_version.py')).read())


@task
def release(ctx, skip_tests=False):
    """
    Release a new version of 'Tailbone'.
    """
    if not skip_tests:
        ctx.run('tox')

    shutil.rmtree('Tailbone.egg-info')
    ctx.run('python setup.py sdist --formats=gztar')
    ctx.run('twine upload dist/Tailbone-{}.tar.gz'.format(__version__))
