
/******************************
 * login.css
 ******************************/

#logo {
    display: block;
    margin: 40px auto;
}

div.form {
    margin: auto;
    float: none;
    text-align: center;
}

div.field-wrapper {
    margin: 10px auto;
    width: 300px;
}

div.field-wrapper label {
    text-align: right;
    width: auto;
}

div.field-wrapper div.field input[type="text"],
div.field-wrapper div.field input[type="password"] {
    margin-left: 1em;
    width: 150px;
}

div.buttons {
    display: block;
}

div.buttons input {
    margin: auto 5px;
}

/* this is for "login as chuck" tip in demo mode */
.tips {
    margin-top: 2em;
    text-align: center;
}
