
/* /\****************************** */
/*  * General */
/*  ******************************\/ */

/* * { */
/*     margin: 0px; */
/* } */

/* body { */
/*     font-family: Verdana, Arial, sans-serif; */
/*     font-size: 11pt; */
/* } */

/* a { */
/*     color: #0972a5; */
/*     text-decoration: none; */
/* } */

/* a:hover { */
/*     text-decoration: underline; */
/* } */

/* h1 { */
/*     margin-bottom: 15px; */
/* } */

/* h2 { */
/*     font-size: 12pt; */
/*     margin: 20px auto 10px auto; */
/* } */

/* li { */
/*     line-height: 2em; */
/* } */

/* p { */
/*     margin-bottom: 5px; */
/* } */

/* .left { */
/*     float: left; */
/*     text-align: left; */
/* } */

/* .right { */
/*     text-align: right; */
/* } */

/* .wrapper { */
/*     overflow: auto; */
/* } */

/* div.buttons { */
/*     clear: both; */
/*     margin-top: 10px; */
/* } */

/* div.dialog { */
/*     display: none; */
/* } */

/* div.flash-message { */
/*     background-color: #dddddd; */
/*     margin-bottom: 8px; */
/*     padding: 3px; */
/* } */

/* div.flash-messages div.ui-state-highlight { */
/*     padding: .3em; */
/*     margin-bottom: 8px; */
/* } */

/* div.error-messages div.ui-state-error { */
/*     padding: .3em; */
/*     margin-bottom: 8px; */
/* } */

/* .flash-messages, */
/* .error-messages { */
/*     margin: 0.5em 0 0 0; */
/* } */

/* ul.error { */
/*     color: #dd6666; */
/*     font-weight: bold; */
/*     padding: 0px; */
/* } */

/* ul.error li { */
/*     list-style-type: none; */
/* } */

/* /\****************************** */
/*  * jQuery UI tweaks */
/*  ******************************\/ */

/* ul.ui-menu { */
/*     max-height: 30em; */
/* } */

/******************************
 * tweaks for root user
 ******************************/

.navbar .navbar-end .navbar-link.root-user,
.navbar .navbar-end .navbar-link.root-user:hover,
.navbar .navbar-end .navbar-link.root-user.is_active,
.navbar .navbar-end .navbar-item.root-user,
.navbar .navbar-end .navbar-item.root-user:hover,
.navbar .navbar-end .navbar-item.root-user.is_active {
    background-color: red;
    font-weight: bold;
}
