
/******************************
 * Grid Filters
 ******************************/

.filters .filter {
    margin-bottom: 0.5rem;
}

.filters .filter-fieldname .field,
.filters .filter-fieldname .field label {
    width: 100%;
}

.filters .filter-fieldname .field label {
    justify-content: left;
}

.filters .filter-verb .select,
.filters .filter-verb .select select {
    width: 100%;
}
