
/**********************************************************************
 * styles for time sheets / schedules
 **********************************************************************/

/******************************
 * header table
 ******************************/

.timesheet-header {
    width: 100%;
}

.timesheet-header td.filters {
    vertical-align: bottom;
    width: 100%;
}

.timesheet-header td.filters .field {
    width: auto;
}

.timesheet-header td.filters .buttons {
    margin-bottom: 0;
}

.timesheet-header td.menu {
    padding: 0.5em;
    vertical-align: top;
    white-space: nowrap;
}

.timesheet-header td.tools {
    margin: 0;
    padding: 0;
    text-align: right;
    vertical-align: bottom;
    white-space: nowrap;
}

.timesheet-header .week-picker label {
    margin-left: 1em;
}

/******************************
 * timesheet table
 ******************************/

.timesheet {
    border-bottom: 1px solid black;
    border-right: 1px solid black;
    clear: both;
    margin: 0.3em 0;
    width: 100%;
}

.timesheet thead th,
.timesheet tbody td {
    border-left: 1px solid black;
    border-top: 1px solid black;
}

.timesheet tbody {
    background-color: white;
}

.timesheet tbody tr:nth-child(odd) {
    background-color: #e0e0e0;
}

.timesheet tbody td {
    padding: 5px;
    text-align: center;
}

.timesheet tbody td.employee {
    text-align: left;
}

.timesheet tbody p.shift {
    display: block;
    font-size: 0.9em;
    margin: 2px;
}

.timesheet tbody tr.total {
    font-weight: bold;
}

/******************************
 * below timesheet...
 ******************************/

div.buttons {
    margin-top: 0;
}
