
/******************************
 * main layout
 ******************************/

body {
    display: flex;
    flex-direction: column;
    min-height: 100vh;
}

.content-wrapper {
    display: flex;
    flex: 1;
    flex-direction: column;
    justify-content: space-between;
}


/******************************
 * header
 ******************************/

header .level {
    /* height: 60px; */
    line-height: 60px;
    padding-left: 0.5em;
    padding-right: 0.5em;
}

header .level #header-logo {
    display: inline-block;
}

header .level .global-title,
header .level-left .global-title {
    font-size: 2em;
    font-weight: bold;
}

header .level #current-context,
header .level-left #current-context {
    font-size: 2em;
    font-weight: bold;
}

header .level #current-context span,
header .level-left #current-context span {
    margin-right: 10px;
}

header .level .theme-picker {
    display: inline-flex;
}

#content-title h1 {
    font-size: 2em;
}

/******************************
 * content
 ******************************/

#page-body {
    padding: 0.4em;
}

/******************************
 * context menu
 ******************************/

#context-menu {
    margin-bottom: 1em;
    margin-left: 1em;
    text-align: right;
    white-space: nowrap;
}

/******************************
 * "object helper" panel
 ******************************/

.object-helper {
    border: 1px solid black;
    margin: 1em;
    padding: 1em;
    width: 20em;
}

.object-helper-content {
    margin-top: 1em;
}

/******************************
 * fix datepicker within modals
 * TODO: someday this may not be necessary? cf.
 * https://github.com/buefy/buefy/issues/292#issuecomment-347365637
 ******************************/

.modal .animation-content .modal-card {
    overflow: visible !important;
}

.modal-card-body {
    overflow: visible !important;
}


/******************************
 * feedback
 ******************************/

.feedback-dialog .red {
    color: red;
    font-weight: bold;
}
