
/******************************
 * main layout
 ******************************/

body {
    display: flex;
    flex-direction: column;
    min-height: 100vh;
}

.content-wrapper {
    display: flex;
    flex: 1;
    flex-direction: column;
    justify-content: space-between;
}


/******************************
 * header
 ******************************/

header .level {
    /* height: 60px; */
    line-height: 60px;
    padding-left: 0.5em;
    padding-right: 0.5em;
}

header .level #header-logo {
    display: inline-block;
}

header .level .global-title,
header .level-left .global-title {
    font-size: 2em;
    font-weight: bold;
}

header .level #current-context,
header .level-left #current-context {
    font-size: 2em;
    font-weight: bold;
}

header .level #current-context span,
header .level-left #current-context span {
    margin-right: 10px;
}

header .level .theme-picker {
    display: inline-flex;
}

/* header .global .grid-nav { */
/*     display: inline-block; */
/*     font-size: 16px; */
/*     font-weight: bold; */
/*     line-height: 60px; */
/*     margin-left: 5em; */
/* } */

/* header .global .grid-nav .ui-button, */
/* header .global .grid-nav span.viewing { */
/*     margin-left: 1em; */
/* } */

#content-title h1 {
    font-size: 2em;
}

/* /\****************************** */
/*  * Logo */
/*  ******************************\/ */

/* #logo { */
/*     display: block; */
/*     margin: 40px auto; */
/* } */


/******************************
 * content
 ******************************/

#page-body {
    padding: 0.4em;
}

/* body > #body-wrapper { */
/*     margin: 0px; */
/*     position: relative; */
/* } */

/* .content-wrapper { */
/*     height: 100%; */
/*     padding-bottom: 30px; */
/* } */

/* #scrollpane { */
/*     height: 100%; */
/* } */

/* #scrollpane .inner-content { */
/*     padding: 0 0.5em 0.5em 0.5em; */
/* } */


/******************************
 * context menu
 ******************************/

#context-menu {
    text-align: right;
    white-space: nowrap;
}

/******************************
 * "object helper" panel
 ******************************/

.object-helper {
    border: 1px solid black;
    margin: 1em;
    padding: 1em;
    min-width: 20em;
}

.object-helper-content {
    margin-top: 1em;
}

/* /\****************************** */
/*  * Panels */
/*  ******************************\/ */

/* .panel-wrapper { */
/*     float: left; */
/*     margin-right: 15px; */
/*     width: 40%; */
/* } */

/* .panel, */
/* .panel-grid { */
/*     border-left: 1px solid Black; */
/*     margin-bottom: 15px; */
/* } */

/* .panel { */
/*     border-bottom: 1px solid Black; */
/*     border-right: 1px solid Black; */
/*     padding: 0px; */
/* } */

/* .panel h2, */
/* .panel-grid h2 { */
/*     border-bottom: 1px solid Black; */
/*     border-top: 1px solid Black; */
/*     padding: 5px; */
/*     margin: 0px; */
/* } */

/* .panel-grid h2 { */
/*     border-right: 1px solid Black; */
/* } */

/* .panel-body { */
/*     overflow: auto; */
/*     padding: 5px; */
/* } */

/******************************
 * feedback
 ******************************/

#feedback-dialog {
    display: none;
}

#feedback-dialog p {
    margin-top: 1em;
}

#feedback-dialog .red {
    color: red;
    font-weight: bold;
}

#feedback-dialog .field-wrapper {
    margin-top: 1em;
    padding: 0;
}

#feedback-dialog .field {
    margin-bottom: 0;
    margin-top: 0.5em;
}

#feedback-dialog .referrer .field {
    clear: both;
    float: none;
    margin-top: 1em;
}

#feedback-dialog textarea {
    width: auto;
}
